/******************************************************************
 * Qiang Zheng,
 * ECS Dept, Syracuse Univ. Nov. 10, 1999
 * for CSE 791 class, fall, 1999, instructor: Steve Chapin
 * O.S.: Linux RedHat 6.0
 * Libraries: Lam MPI library and pnm library for image processing
 ******************************************************************/

#include <stdio.h>
#include <string.h>
#include <mpi.h>
#include <pnm.h>
#include <math.h>
#include <stdlib.h>

/*
 * Data types will be used which are defined in PNM lib:
 * typedef ... xel;              three xelvals
 * typedef ... xelval;           range from 0 to PNM_MAXMAXVAL
 * #define PNM_MAXMAXVAL ...     normally it is 1
 * extern xelval pnm_pbmmaxval;
 *
 * Input and Output functions available in pnm library:
 * Input: xel** pnm_readpnm( FILE* fp, int* colsP, int* rowsP
 *                           , xelval* maxvalP, int* formatP)
 * Output: void pnm_writepnm( FILE* fp, xel** xels, int cols
 *                            , int rows, xelval maxval, int format
 *                            , int forceplain )
 */

#define	WORKTAG	1
#define DIETAG	0

#define UPLIMIT 10
#define LOLIMIT 0

#define MAXLINE 100
/* the max length of a line */

main( int argc, char *argv[] )
{
  int myRank;

  MPI_Init( &argc, &argv ); /* Initialize MPI environment */

  MPI_Comm_rank( MPI_COMM_WORLD, &myRank ); 

  if( myRank == 0 ){
      master( argc, argv );         /* do the master job */
  } else {
      slave( myRank );              /* do the slave job */
  }

  MPI_Finalize();                   /* cleanup MPI */
}

/*
 * The master function:
 * 1) reads image files,
 * 2) partition the image matrix into blocks,
 * 3) assign them to different slave processes,
 * 4) gather sub-blocks computed by slaves
 * 5) and write the result to image files
 * The schema is map one or several rows into one process
 * so that the boundaries between two adjecent blocks are just
 * two rows, the last row of the previous block and the first
 * row of the next block.
 */
master( int argc, char* argv[] )
{
    /* image stuffs */
    char *oldFname, *newFname, *p, line[ MAXLINE ]; 
    FILE *origF = NULL, *destF=NULL;
    int colNum, rowNum, colCount, rowCount, blockSize, format;
    int rowsPerProc, rowsInLastBlock, lastBlockSize, i, j;
    xel **image, **newImage, **temp0, **temp1;
    xel *oldXel, *newXel, *lastRow;
    /* since images are matrices, they can be regarded as rectangular shape */
    xelval maxval, pVal;
    pixval gVal, rVal, bVal;

    int beginRow; /* debug variable */

    /* MPI stuffs */
    int iterTimes=10, ntasks, rank;

    MPI_Status  status;

    while ( 1 ) {
      printf("Please input the file name(enter 'q' to quit):\n");
      scanf( "%s",line );
      if( !strcmp( line, "q" ) ) break;

      pnm_init( &argc, argv ); /* routine function in pnm library */

      origF = fopen( line, "r");

      if ( origF==NULL ) {
        printf( "No such file %s or permission denied. \n", argv[ 1 ] );
        exit( 1 );
      }

      printf("Please input the iteration times (integer,10 by default):\n");
      scanf("%d", &iterTimes);

      printf("Smooth %d times, a moment please ...\n", iterTimes);

      /* making the file name */
      oldFname = line;
      newFname = ( char * ) calloc( strlen( oldFname )+3, sizeof( char ) );
      p = newFname;
      while( *oldFname != '.' ) {
        *p++ = *oldFname++;
      }
      *p++ = '_'; *p++ = 's';
      while( *oldFname != '\0' ) {
        *p++ = *oldFname++;
      }

      *p++ = '\0';

      destF = fopen( newFname, "w");

      MPI_Comm_size( MPI_COMM_WORLD, &ntasks );

printf("Master: number of nodes = %d \n", ntasks);

      image = pnm_readpnm( origF, &colNum, &rowNum, &maxval, &format);
      /* newImage = ( xel ** )calloc( rowNum, sizeof( xel* ) ); */
      newImage=image;

      lastRow = image[ rowNum-1 ]; /* offset of the last row */

      /* master doesn't calculate */
      rowsPerProc = ceil( ( double )rowNum/( double )(ntasks-1) );
      rowsInLastBlock = rowNum-rowsPerProc*(ntasks-2);
      /* size of the block of data each slave process gets */
      blockSize = colNum*rowsPerProc;
      /* size of the last data block the last slave process gets */
      lastBlockSize = colNum * rowsInLastBlock;

      for( rank=1; rank<ntasks; rank++ ){

        /* 
         * prototype :
         * MPI_Send( buffer, count, datatype, destination, tag, MPI_COM_WORLD );
         */

        /* To tell the slave how many iteration times are needed */
        MPI_Send( &iterTimes, 1, MPI_INT, rank, WORKTAG, MPI_COMM_WORLD );
        MPI_Send( &format, 1, MPI_INT, rank, WORKTAG, MPI_COMM_WORLD );
        /* To tell the slave how many columns in each row */
        MPI_Send( &colNum, 1, MPI_INT, rank, WORKTAG, MPI_COMM_WORLD );

        /* To tell slaves the size of the blocks will be sent to them */
        if( rank==ntasks-1 ) {
          MPI_Send( &rowsInLastBlock, 1, MPI_INT, rank, WORKTAG, MPI_COMM_WORLD );
        } else {
          MPI_Send( &rowsPerProc, 1, MPI_INT, rank, WORKTAG, MPI_COMM_WORLD );
        }

        printf("Sending data to slaves...\n");

        /*
         * To send the data block to slaves
         * if it is the last slave, the block size may be less than
         * the ragular block size
         */
        if ( rank==ntasks-1 ) {
          for( i=0, temp0=image+(rank-1)*rowsPerProc;
               i<rowsInLastBlock;
               i++, temp0++ ) {
               MPI_Send( *temp0, colNum*sizeof(xel), MPI_BYTE, rank,
                 WORKTAG, MPI_COMM_WORLD );
               /* To send one row of data each time */
          }
        } else {
          for(i=0, temp0=image+(rank-1)*rowsPerProc;
              i<rowsPerProc;
              i++, temp0++) {
            MPI_Send( *temp0, colNum*sizeof(xel), MPI_BYTE, rank,
              WORKTAG, MPI_COMM_WORLD );
	}
        }
       }

      printf("Receiving computed result from slaves \n");
      for( rank=1 ; rank<ntasks-1; rank++ ) {

          for(i=0, temp1=newImage+(rank-1)*rowsPerProc;
              i<rowsPerProc; i++, temp1++) {
            /* To receive one row of data each time */

          /* *temp1 = ( xel * ) calloc(colNum, sizeof( xel ) ); */
            MPI_Recv( *temp1, colNum*sizeof(xel), MPI_BYTE, rank,
              MPI_ANY_TAG, MPI_COMM_WORLD , &status);
          }
      } 

     /* receive data from last node */
     for( i=0, temp1=newImage+(ntasks-2)*rowsPerProc;
          i<rowsInLastBlock; i++, temp1++ ) {
          /* To receive one row of data each time */
          /* *temp1 = ( xel * ) calloc(colNum, sizeof( xel ) ); */
          MPI_Recv( *temp1, colNum*sizeof(xel), MPI_BYTE, ntasks-1, 
              MPI_ANY_TAG, MPI_COMM_WORLD, &status );
     } 
          /* re-organize the data collected from slaves */

          /* MPI_Send( 0, 0, MPI_INT, rank, DIETAG, MPI_COMM_WORLD ); */

    pnm_writepnm( destF, newImage, colNum, rowNum, maxval, format, 0 ); 

    /* free memory */
    /* pnm_freearray( image, rowNum );
    pnm_freearray( newImage, rowNum ); */

    printf("Computing finished ! \n");

  }
}

/*
 * the slave funtion smooth the block it gets from the master
 * and return the result, if necessary, it will exchange the boundrary
 * data with its neighbors
 */ 
slave( int rankNum )
{
    int         blockSize = 0, iterTimes, colNum, rowNum, i, j, k,
                myRank, ntasks, format,
                prevRank, nextRank; /* rank of last and next nodes */
    xel         *work, *swapBlock, *temp0, *temp1, *upBound, *lowBound,
                *prevRow, *nextRow, *lastRow, *tempP;
    xelval      north, south, west, east, center;
    pixval      gNorth, gSouth, gWest, gEast, gCenter,
                rNorth, rSouth, rWest, rEast, rCenter,
                bNorth, bSouth, bWest, bEast, bCenter;
    /* pointer `work' points to the block of data to be computed
     * including the two boundaries,
     * swapBlock points to another copy
     * upBound points to the first row of work buffer, e.g. upper boundary
     * lowBound points to the last row of work buffer, e.g. lower boundary
     */
    MPI_Status  status;

    myRank = rankNum;
    MPI_Comm_size( MPI_COMM_WORLD, &ntasks );
    nextRank = myRank%(ntasks-1)+1;
    prevRank = ( myRank+( ntasks-3 ) )%( ntasks-1 )+1;

/* printf("node %d next node is %d\n", myRank, nextRank);
printf("node %d previous node is %d\n", myRank,prevRank); */

    /* to receive the iteration times, column number sent to it */
    MPI_Recv( &iterTimes, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
    MPI_Recv( &format, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
    MPI_Recv( &colNum, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
    /* to receive the number of rows to be sent to this node */
    MPI_Recv( &rowNum, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );

    blockSize=rowNum*colNum;

    /*
    /* Allocate memory for the data block and a swap block it will receive */
    work = ( xel * ) calloc( blockSize, sizeof( xel ) ); 
    swapBlock = ( xel * ) calloc( blockSize, sizeof( xel ) ); 
    tempP = work;
    lastRow = work+blockSize-colNum; 

    /* to receive the data row by row */
    for ( i=0, temp0=work; i<rowNum; i++, temp0+=colNum ) {
      MPI_Recv( temp0, colNum*sizeof( xel ), MPI_BYTE,
        0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
    }

    upBound = ( xel * ) calloc( colNum, sizeof( xel ) );
    lowBound = ( xel * ) calloc( colNum, sizeof( xel ) );

    for ( k=0; k<iterTimes; k++ ) {

      printf("Iteration %d in node %d \n", k, myRank );
      /* Before computing begins, exchange data with
       * neighborhood
       */

      /* Exchanging data with its neighbours */

      MPI_Send( work, colNum*sizeof( xel ), MPI_BYTE,
        prevRank, WORKTAG, MPI_COMM_WORLD ); 
      MPI_Send( work+blockSize-colNum, colNum*sizeof( xel ), MPI_BYTE,
        nextRank, WORKTAG, MPI_COMM_WORLD );  

      MPI_Recv( lowBound, colNum*sizeof( xel ), MPI_BYTE,
        nextRank, MPI_ANY_TAG, MPI_COMM_WORLD, &status ); 
      MPI_Recv( upBound, colNum*sizeof( xel ), MPI_BYTE,
        prevRank, MPI_ANY_TAG, MPI_COMM_WORLD, &status ); 

      if( myRank==1 ) {
        memcpy( upBound, work, colNum*sizeof( xel ) );
      }

      if( myRank== ntasks-1 ) {
	memcpy( lowBound, work+blockSize-colNum, colNum*sizeof( xel ) ); 
      }

      /* computing begins here */

      prevRow = upBound; nextRow=work+colNum;
      temp0=work; temp1=swapBlock;

      switch ( PNM_FORMAT_TYPE( format ) ) {
        case PPM_TYPE:
          for( i=0; i<rowNum-1; i++ ) {
            for( j=0; j<colNum; j++ ) {
              gCenter = PPM_GETG( temp0[ j ] );
              gSouth = PPM_GETG( nextRow[ j ] );
              gNorth = PPM_GETG( prevRow[ j ] );

              rCenter = PPM_GETR( temp0[ j ] );
              rSouth = PPM_GETR( nextRow[ j ] );
              rNorth = PPM_GETR( prevRow[ j ] );

              bCenter = PPM_GETB( temp0[ j ] );
              bSouth = PPM_GETB( nextRow[ j ] );
              bNorth = PPM_GETB( prevRow[ j ] );

              if( j==0 ) { /* left boundary */
                gEast = PPM_GETG( temp0[ j+1 ] );
                gWest = PPM_GETG( temp0[ j ] );

                rEast = PPM_GETR( temp0[ j+1 ] );
                rWest = PPM_GETR( temp0[ j ] );

                bEast = PPM_GETB( temp0[ j+1 ] );
                bWest = PPM_GETB( temp0[ j ] );

              } else if (j==colNum-1 ) { /* right boundary */
                gEast = PPM_GETG( temp0[ j ] );
                gWest = PPM_GETG( temp0[ j-1 ] );

                rEast = PPM_GETR( temp0[ j ] );
                rWest = PPM_GETR( temp0[ j-1 ] );
                
                bEast = PPM_GETB( temp0[ j ] );
                bWest = PPM_GETB( temp0[ j-1 ] );
              
              } else { /* Non-boundary pixels */
                gEast = PPM_GETG( temp0[ j+1 ] );
                gWest = PPM_GETG( temp0[ j-1 ] );

                rEast = PPM_GETR( temp0[ j+1 ] );
                rWest = PPM_GETR( temp0[ j-1 ] );
                
                bEast = PPM_GETB( temp0[ j+1 ] );
                bWest = PPM_GETB( temp0[ j-1 ] );
              
	      }

              gCenter = (gSouth+gNorth+gEast+gWest+gCenter)/5;
              rCenter = (rSouth+rNorth+rEast+rWest+rCenter)/5;
              bCenter = (bSouth+bNorth+bEast+bWest+bCenter)/5; 
              PPM_ASSIGN( temp1[ i*colNum+j ], rCenter, gCenter, bCenter );

            }
            prevRow = temp0; temp0=nextRow; 
  
            if(temp0<work+blockSize-colNum)
              nextRow = nextRow+colNum;
            else
              nextRow = lowBound;
          }
          break;

        case PGM_TYPE:
        case PBM_TYPE:
          for( i=0; i<rowNum-1; i++ ) {
            for( j=0; j<colNum; j++ ) {
              center = PNM_GET1( temp0[ j ] );
              south = PNM_GET1( nextRow[ j ] );
              north = PNM_GET1( prevRow[ j ] );

              if( j==0 ) { /* left boundary */
                east = PNM_GET1( temp0[ j+1 ] );
                west = PNM_GET1( temp0[ j ] );
              } else if ( j==colNum-1 ) { /* right boundary */
                east = PNM_GET1( temp0[ j ] );
                west = PNM_GET1( temp0[ j-1 ] );
              } else { /* Non-boundary pixels */
                east = PNM_GET1( temp0[ j+1 ] );
                west = PNM_GET1( temp0[ j-1 ] );
              }
              center = (south+north+east+west+center)/5; 

              /* *temp1 = ( xel * ) calloc(colNum, sizeof( xel ) ); */
              PNM_ASSIGN1( temp1[ i*colNum+j ], center );
            }
          prevRow = temp0; temp0=nextRow;
  
          if(temp0<work+blockSize-colNum)
            nextRow = nextRow+colNum;
          else
            nextRow = lowBound;
        }
        break;
      }
        /* swap the pointers to original buffer and
	 * newly computed result buffer */
	tempP = work;
        work = swapBlock;
        swapBlock = tempP;
      }

    for( i=0, tempP=work; i<rowNum; i++, tempP+=colNum ) {
      MPI_Send( tempP, colNum*sizeof(xel), MPI_BYTE, 0, 0, MPI_COMM_WORLD );
      /* send result row by row */
    }

    /* if( status.MPI_TAG==DIETAG ) {
      return;
    }*/
/* return; */

/* free memory */
}
